/*
 *                      Yeppp! library implementation
 *
 * This file is part of Yeppp! library and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

#include <jni.h>
#include <yepLibrary.h>

#if defined(YEP_WINDOWS_OS)
	#include <windows.h>

	#if defined(YEP_CUSTOM_RUNTIME)
		BOOL WINAPI _DllMainCRTStartup(HINSTANCE instance, DWORD reason, LPVOID reserved) {
			return TRUE;
		}
	#else
		BOOL WINAPI DllMain(HINSTANCE instance, DWORD reason, LPVOID reserved) {
			return TRUE;
		}
	#endif
#endif

YEP_PRIVATE_SYMBOL jclass NullPointerException = NULL;
YEP_PRIVATE_SYMBOL jclass IndexOutOfBoundsException = NULL;
YEP_PRIVATE_SYMBOL jclass IllegalArgumentException = NULL;
YEP_PRIVATE_SYMBOL jclass UnsupportedOperationException = NULL;

static void cleanup(JNIEnv *env) {
	if (NullPointerException) {
		(*env)->DeleteGlobalRef(env, NullPointerException);
	}
	if (IndexOutOfBoundsException) {
		(*env)->DeleteGlobalRef(env, IndexOutOfBoundsException);
	}
	if (IllegalArgumentException) {
		(*env)->DeleteGlobalRef(env, IllegalArgumentException);
	}
	if (UnsupportedOperationException) {
		(*env)->DeleteGlobalRef(env, UnsupportedOperationException);
	}
}

JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *jvm, void *reserved) {
	JNIEnv *env;
	enum YepStatus status;
	jint jnicallResult;
	jclass localNullPointerException = NULL, localIllegalArgumentException = NULL, localIndexOutOfBoundsException = NULL, localUnsupportedOperationException = NULL;

	jnicallResult = (*jvm)->GetEnv(jvm, (void**)&env, JNI_VERSION_1_6);
	if (jnicallResult != JNI_OK) {
		return JNI_ERR;
	}

	localNullPointerException = (*env)->FindClass(env, "java/lang/NullPointerException");
	if (localNullPointerException) {
		NullPointerException = (jclass) (*env)->NewGlobalRef(env, localNullPointerException);
		(*env)->DeleteLocalRef(env, localNullPointerException);
	} else {
		cleanup(env);
		return JNI_ERR;
	}

	localIndexOutOfBoundsException = (*env)->FindClass(env, "java/lang/IndexOutOfBoundsException");
	if (localIndexOutOfBoundsException) {
		IndexOutOfBoundsException = (jclass) (*env)->NewGlobalRef(env, localIndexOutOfBoundsException);
		(*env)->DeleteLocalRef(env, localIndexOutOfBoundsException);
	} else {
		cleanup(env);
		return JNI_ERR;
	}

	localIllegalArgumentException = (*env)->FindClass(env, "java/lang/IllegalArgumentException");
	if (localIllegalArgumentException) {
		IllegalArgumentException = (jclass) (*env)->NewGlobalRef(env, localIllegalArgumentException);
		(*env)->DeleteLocalRef(env, localIllegalArgumentException);
	} else {
		cleanup(env);
		return JNI_ERR;
	}

	localUnsupportedOperationException = (*env)->FindClass(env, "java/lang/UnsupportedOperationException");
	if (localUnsupportedOperationException) {
		UnsupportedOperationException = (jclass) (*env)->NewGlobalRef(env, localUnsupportedOperationException);
		(*env)->DeleteLocalRef(env, localUnsupportedOperationException);
	} else {
		cleanup(env);
		return JNI_ERR;
	}

	status = yepLibrary_Init();
	if (status == YepStatusOk) {
		return JNI_VERSION_1_6;
	} else {
		return JNI_ERR;
	}
}

JNIEXPORT void JNICALL JNI_OnUnload(JavaVM *jvm, void *reserved) {
	JNIEnv *env;
	jint jnicallResult;

	yepLibrary_Release();

	jnicallResult = (*jvm)->GetEnv(jvm, (void**)&env, JNI_VERSION_1_6);
	if (jnicallResult != JNI_OK) {
		return;
	}

	cleanup(env);
}

JNIEXPORT jlong JNICALL Java_info_yeppp_Library_getCpuIsaFeatures(JNIEnv *env, jclass class) {
	Yep64u isaFeatures = YepIsaFeaturesDefault;
	enum YepStatus status;
	status = yepLibrary_GetCpuIsaFeatures(&isaFeatures);
	return isaFeatures;
}

JNIEXPORT jlong JNICALL Java_info_yeppp_Library_getCpuSimdFeatures(JNIEnv *env, jclass class) {
	Yep64u simdFeatures = YepSimdFeaturesDefault;
	enum YepStatus status;
	status = yepLibrary_GetCpuSimdFeatures(&simdFeatures);
	return simdFeatures;
}

JNIEXPORT jlong JNICALL Java_info_yeppp_Library_getCpuSystemFeatures(JNIEnv *env, jclass class) {
	Yep64u systemFeatures = YepSystemFeaturesDefault;
	enum YepStatus status;
	status = yepLibrary_GetCpuSimdFeatures(&systemFeatures);
	return systemFeatures;
}

JNIEXPORT jlong JNICALL Java_info_yeppp_Library_getCpuSystemFeatures(JNIEnv *env, jclass class) {
	Yep64u systemFeatures = YepSystemFeaturesDefault;
	enum YepStatus status;
	status = yepLibrary_GetCpuSimdFeatures(&systemFeatures);
	return systemFeatures;
}

JNIEXPORT jint JNICALL Java_info_yeppp_Library_getCpuVendorId(JNIEnv *env, jclass class) {
	enum YepCpuVendor vendor = YepCpuVendorUnknown;
	enum YepStatus status;
	status = yepLibrary_GetCpuVendor(&vendor);
	return vendor;
}

JNIEXPORT jint JNICALL Java_info_yeppp_Library_getCpuArchitectureId(JNIEnv *env, jclass class) {
	enum YepCpuArchitecture architecture = YepCpuArchitectureUnknown;
	enum YepStatus status;
	status = yepLibrary_GetCpuVendor(&architecture);
	return architecture;
}

JNIEXPORT jint JNICALL Java_info_yeppp_Library_getCpuMicroarchitectureId(JNIEnv *env, jclass class) {
	enum YepCpuMicroarchitecture microarchitecture = YepCpuMicroarchitecture;
	enum YepStatus status;
	status = yepLibrary_GetCpuVendor(&microarchitecture);
	return microarchitecture;
}

JNIEXPORT jlong JNICALL Java_info_yeppp_Library_getTimerTicks(JNIEnv *env, jclass class) {
	Yep64u ticks = 0ull;
	enum YepStatus status;
	status = yepLibrary_GetTimerTicks(&ticks);
	return ticks;
}

JNIEXPORT jlong JNICALL Java_info_yeppp_Library_getTimerFrequency(JNIEnv *env, jclass class) {
	Yep64u frequency = 0ull;
	enum YepStatus status;
	status = yepLibrary_GetTimerFrequency(&frequency);
	return frequency;
}

JNIEXPORT jlong JNICALL Java_info_yeppp_Library_getTimerAccuracy(JNIEnv *env, jclass class) {
	Yep64u accuracy = 0ull;
	enum YepStatus status;
	status = yepLibrary_GetTimerAccuracy(&accuracy);
	return accuracy;
}
