/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build;

import info.yeppp.Build.internals.CommandLineTool;
import info.yeppp.Build.internals.FileSystem;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.SystemUtils;

import java.io.File;
import java.io.IOException;
import java.util.*;

public class GnuCppCompiler extends AbstractCppCompiler {
	static enum PositionIndependentCodeGeneration {
		NoPIC,
		LimitedLibraryPIC,
		UnlimitedLibraryPIC,
		LimitedExecutablePIC,
		UnlimitedExecutablePIC
	}

	public GnuCppCompiler() {
	}

	public BuildMessages compile(String sourcePath) {
		if (sourceDirectory == null) {
			throw new IllegalStateException("Source directory was not assigned any value");
		}
		if (objectDirectory == null) {
			throw new IllegalStateException("Object directory was not assigned any value");
		}
		final BuildMessages buildMessages = new BuildMessages();
		final String relativeSourcePath = FileSystem.getRelativePath(sourcePath, sourceDirectory);
		if (relativeSourcePath == null) {
			throw new IllegalArgumentException("Source path is not within source directory");
		}
		final String objectPath = FilenameUtils.concat(objectDirectory,
				FilenameUtils.removeExtension(relativeSourcePath) + (SystemUtils.IS_OS_WINDOWS ? ".obj" : ".o"));

		createDirectories(objectPath);
        FileUtils.deleteQuietly(new File(objectPath));
		try {
			final List<String> commandLine = composeCommandLine(sourcePath, objectPath);
			final Map<String, String> environment = composeEnvironment();
            System.out.println(StringUtils.join(commandLine, " "));
			final CommandLineTool compiler = new CommandLineTool(commandLine, environment);
			for (String line : compiler.getError()) {
				System.err.println(line);
			}
			List<String> output = compiler.getOutput();
			File sourceFile = new File(sourcePath);
			for (String line : compiler.getOutput()) {
				System.out.println(line);
			}

		} catch (IOException exception) {
			buildMessages.add(new BuildMessage(BuildMessage.Type.Exception, relativeSourcePath, null, null, exception.getMessage()));
		} catch (Exception e) {
			e.printStackTrace();
		}
		return buildMessages;
	}

	protected List<String> composeCommandLine(String sourcePath, String objectPath) {
		final List<String> commandLine = new ArrayList<String>();
		commandLine.add("g++");
		commandLine.add("-c");
		commandLine.add("-g");
		commandLine.add("-o" + objectPath);
		for (String include : includes) {
			commandLine.add("-I" + include);
		}
		switch (optimization) {
			case NoOptimization:
				commandLine.add("-O0"); break;
			case MinSizeOptimization:
				commandLine.add("-Os"); break;
			case SpeedOptimization:
				commandLine.add("-O2"); break;
			case MaxSpeedOptimization:
				commandLine.add("-O3"); break;
		}
		for (Map.Entry<String, String> macro: macros.entrySet()) {
			commandLine.add("-D" + macro.getKey() + "=" + macro.getValue());
		}
		commandLine.add("-Wa,--noexecstack");
		switch (runtimeLibrary) {
			case NoRuntimeLibrary:
				commandLine.add("-nostdlib"); break;
			case DynamicRuntimeLibrary:
				commandLine.add("-shared-libgcc"); break;
			case StaticRuntimeLibrary:
				commandLine.add("-static-libgcc"); break;
		}
		commandLine.add(rttiEnabled ? "-frtti" : "-fno-rtti");
		switch (exceptions) {
			case NoExceptions:
				commandLine.add("-fno-exceptions");
				commandLine.add("-fno-unwind-tables");
				break;
			case DisabledExceptions:
				commandLine.add("-fno-exceptions");
				commandLine.add("-funwind-tables");
				break;
			case SyncronousExceptions:
				commandLine.add("-fexceptions");
				commandLine.add("-funwind-tables");
				break;
			case AsyncronousExceptions:
				commandLine.add("-fexceptions");
				commandLine.add("-funwind-tables");
				commandLine.add("-fasynchronous-unwind-tables");
				commandLine.add("-fnon-call-exceptions");
				break;
		}
		switch (pic) {
			case NoPIC:
				break;
			case LimitedLibraryPIC:
				commandLine.add("-fpic");
				break;
			case UnlimitedLibraryPIC:
				commandLine.add("-fPIC");
				break;
			case LimitedExecutablePIC:
				commandLine.add("-fpie");
				break;
			case UnlimitedExecutablePIC:
				commandLine.add("-fPIE");
				break;
		}
		if (runtimeLibrary.equals(RuntimeLibrary.NoRuntimeLibrary)) {

		}
		commandLine.add(sourcePath);
		return commandLine;
	}

	private Map<String, String> composeEnvironment() throws IOException {
		final Map<String, String> environment = new HashMap<String, String>(System.getenv());
		boolean processedIncludeVariable = false;
		for (Map.Entry<String, String> variable : environment.entrySet()) {
			if (variable.getKey().equalsIgnoreCase("INCLUDE")) {
				String include = variable.getValue();
				include = StringUtils.join(globalIncludes, SystemUtils.PATH_SEPARATOR) +
						SystemUtils.PATH_SEPARATOR + include;
				variable.setValue(include);
				processedIncludeVariable = true;
			}
		}
		if (processedIncludeVariable == false) {
			environment.put("INCLUDE", StringUtils.join(globalIncludes, SystemUtils.PATH_SEPARATOR));
		}
		return environment;
	}

	private void createDirectories(String path) {
		File file = new File(path);
		if (!file.isDirectory()) {
			file = file.getParentFile();
		}
		file.mkdirs();
	}

	@Override
	public void addDefaultGlobalIncludeDirectories() {
	}

	public void setPositionIndependentCodeGeneration(PositionIndependentCodeGeneration pic) {
		this.pic = pic;
	}

	public PositionIndependentCodeGeneration getPositionIndependentCodeGeneration() {
		return this.pic;
	}

	protected PositionIndependentCodeGeneration pic = PositionIndependentCodeGeneration.NoPIC;
}
